#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : tql-Python.
# @File         : demo
# @Time         : 2019-09-23 13:28
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 


import time
import numpy as np
from sklearn.datasets import make_classification
from sklearn.model_selection import KFold
from tql.algo_ml.cv import LGBMClassifierCV
import os
import socket

if socket.gethostname() == 'yuanjie-Mac.local':
    os.environ['KMP_DUPLICATE_LIB_OK'] = 'TRUE'


class UnderSamplerEnsemble(object):

    def __init__(self, params=None, cv=5, random_state=None):
        self.oof = LGBMClassifierCV(params, cv, random_state)

    def fit(self, X_p, X_n, X_test, cv=5, random_state=666, pred_ranking=False):
        kf = KFold(cv, True, random_state)

        self.preds = []
        for n_fold, (_, idx) in enumerate(kf.split(X_n)):
            print(f"UnderSamplerEnsemble {n_fold + 1} started at {time.ctime()}")
            X = np.r_[X_n[idx], X_p]
            y = np.r_[[0] * len(idx), [1] * len(X_p)]

            self.oof.fit(X, y, X_test)
            self.preds.append(self.oof.oof_test_rank if pred_ranking else self.oof.oof_test)

        return np.column_stack(self.preds).mean(1)


if __name__ == '__main__':
    from sklearn.metrics import roc_auc_score

    X, y = make_classification(
        n_classes=2, class_sep=1.5, weights=[0.999, 0.001],
        n_informative=3, n_redundant=1, flip_y=0,
        n_features=20, n_clusters_per_class=1,
        n_samples=10000, random_state=10
    )

    X_p = X[y == 1]
    X_n = X[y == 0]

    use = UnderSamplerEnsemble()
    preds = use.fit(X_p, X_n, X)
    print(roc_auc_score(y, preds))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : tql-Python.
# @File         : __init__.py
# @Time         : 2019-06-23 20:07
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

from .LogisticRegressionCV import LogisticRegressionCV
from .LGBMClassifierCV import LGBMClassifierCV
from .XGBClassifierCV import XGBClassifierCV
from .CatBoostClassifierCV import CatBoostClassifierCV
from .KerasCV import KerasCV

# TODO: 各种默认参数



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : tql-Python.
# @File         : LogisticRegressionCV
# @Time         : 2019-09-17 22:33
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

import time
import numpy as np
import pandas as pd
from sklearn.linear_model import LogisticRegression, SGDClassifier
from sklearn.model_selection import StratifiedKFold, RepeatedStratifiedKFold
from sklearn.metrics import roc_auc_score
from scipy.sparse.csr import csr_matrix


class LogisticRegressionCV(object):

    def __init__(self, cv=5, random_state=None, n_repeats=None):
        self.clf = LogisticRegression(random_state=random_state, n_jobs=-1)

        if n_repeats:
            self._kf = RepeatedStratifiedKFold(cv, True, random_state)
            self._num_preds = cv * n_repeats
        else:
            self._kf = StratifiedKFold(cv, True, random_state)
            self._num_preds = cv

    def fit(self, X, y, X_test=None, feval=roc_auc_score, sample_weight=None):
        """输入数组"""
        if X_test is None:
            X_test = X[:1]

        self.oof_train = np.zeros(len(X) if not isinstance(X, csr_matrix) else X.shape[0])
        self.oof_test = np.zeros((len(X_test) if not isinstance(X, csr_matrix) else X_test.shape[0], self._num_preds))
        for n_fold, (train_index, valid_index) in enumerate(self._kf.split(X, y)):
            print("\033[94mFold %s started at %s\033[0m" % (n_fold + 1, time.ctime()))
            X_train, y_train = X[train_index], y[train_index]
            X_valid, y_valid = X[valid_index], y[valid_index]
            eval_set = [(X_train, y_train), (X_valid, y_valid)]

            ########################################################################
            self.clf.fit(X_train, y_train, sample_weight=None)

            self.oof_train[valid_index] = self.clf.predict_proba(X_valid)[:, 1]
            self.oof_test[:, n_fold] = self.clf.predict_proba(X_test)[:, 1]
            ########################################################################

        # 输出 测试集 oof
        self.oof_test_rank = (pd.DataFrame(self.oof_test).rank().mean(1) / len(self.oof_test)).values
        self.oof_test = self.oof_test.mean(1)

        # 计算 训练集 oof 得分
        if feval:
            self.oof_score = feval(y, self.oof_train)
            print(f"\n\033[94mCV Score: {self.oof_score} ended at {time.ctime()}\033[0m")
            return self.oof_score



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
__title__ = '__init__.py'
__author__ = 'JieYuan'
__mtime__ = '19-2-18'
"""

from .Optimizer import Optimizer
from .LGBMOptimizer import LGBMOptimizer
from .XGBOptimizer import XGBOptimizer
from .CatBoostOptimizer import CatBoostOptimizer



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
__title__ = 'Optimizer'
__author__ = 'JieYuan'
__mtime__ = '19-3-18'
"""
import numpy as np
from bayes_opt import BayesianOptimization

from ..cv import LGBMClassifierCV


class Optimizer(object):
    """默认LGBMClassifierCV"""

    def __init__(self, X, y, cv=5, cv_seed=0, params_bounds=None):
        """https://www.cnblogs.com/wzdLY/p/9867719.html
        """
        self.X = X
        self.y = y
        self.cv = cv
        self.cv_seed = cv_seed

        self.params_bounds = params_bounds if params_bounds else \
            dict(
                # cat_smooth
                metric='auc',
                class_weight=None,  # 'balanced'
                boosting_type='gbdt',
                num_leaves=(2 ** 2, 2 ** 7),
                learning_rate=(0.001, 1),
                min_split_gain=(0.001, 1),
                min_child_weight=(0.001, 64),  # 终点节点最小样本占比的和 TODO 确定范围
                min_child_samples=(16, 128),
                subsample=(0.1, 1),
                colsample_bytree=(0.1, 1),
                subsample_freq=(2, 6),
                reg_alpha=(0, 6),
                reg_lambda=(0, 6),
                n_jobs=-1,
                importance_type='split',  # 'gain'
                verbosity=-1,

                random_state=666,
                bagging_seed=0,
                feature_fraction_seed=0
            )

    def maximize(self, n_iter=3, seed=666, return_best_params=True):
        self.optimizer = BayesianOptimization(
            f=self.objective,
            pbounds={k: v for k, v in self.params_bounds.items() if isinstance(v, tuple)},
            random_state=seed,
            verbose=2
        )
        self.optimizer.maximize(init_points=5, n_iter=n_iter)

        if return_best_params:
            _ = self.optimizer.max
            print(f"Max Score: {_['target']}")
            return self._round_params({**self.params_bounds, **_['params']})

    def objective(self, **params):
        """重写目标函数"""

        # 纠正参数类型
        params = self._round_params(params)
        _params = {**self.params_bounds, **params}
        _params['n_estimators'] = 10000

        # 核心逻辑
        self.clf = LGBMClassifierCV(_params, self.cv, self.cv_seed)
        return self.clf.fit(self.X, self.y, early_stopping_rounds=300, verbose=0)

    def _round_params(self, params):
        for p in ('max_depth', 'depth', 'num_leaves', 'subsample_freq', 'min_child_samples'):
            if p in params:
                params[p] = int(np.round(params[p]))
        return params


if __name__ == '__main__':
    from sklearn.datasets import make_classification

    X, y = make_classification(10000)
    opt = Optimizer(X, y)
    print(opt.maximize(1))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
__title__ = 'read2write'
__author__ = 'JieYuan'
__mtime__ = '19-3-15'
"""

from tqdm import tqdm


def read2write(ifile, ofile, progress_func=lambda x: x, lower=True, overwrite=True):
    with open(ifile, 'r', encoding='utf8') as input, \
            open(ofile, 'w' if overwrite else 'a', encoding='utf8') as output:
        for line in tqdm(input):
            if line.strip():
                _ = progress_func(line.lower() if lower else line)
                output.write(_ + '\n')


if __name__ == '__main__':
    read2write('/Users/yuanjie/Desktop/notebook/0_MI/training/17/part-00000',
               '/Users/yuanjie/Desktop/notebook/0_MI/test.json',
               lambda x: str(eval(x)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : tql-Python.
# @File         : XGBOptimizer
# @Time         : 2019-09-16 11:58
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

import json
import yaml
import numpy as np
import xgboost as xgb

from .Optimizer import Optimizer


class XGBOptimizer(Optimizer):

    def __init__(self, X, y, cv=5, cv_seed=0, params_bounds=None, use_gpu=1):
        super().__init__(X, y, cv, cv_seed, params_bounds)

        self.best_iterations = []
        self.data = xgb.DMatrix(X, y, silent=False, nthread=8)
        self.params_bounds = params_bounds if params_bounds else \
            dict(
                eval_metric='auc',
                booster='gbtree',
                objective='binary:logistic',

                max_depth=(3, 16),
                learning_rate=(0.001, 1),
                max_delta_step=(0, 2),  # 极端不平衡的时候有用

                gamma=(0.001, 1),  # 描述分裂的最小 gain, 控制树的有用的分裂
                min_child_weight=(0.001, 100),  # 决定最小叶子节点样本权重和,使一个结点分裂的最小权值之和, 避免过拟合

                subsample=(0.1, 1),
                colsample_bytree=(0.1, 1),  # 每棵树的列数
                # colsample_bylevel=(0.3, 1),  # 每一层的列数

                reg_alpha=(0, 10),
                reg_lambda=(0, 10),

                scale_pos_weight=1,

                random_state=666,
                n_jobs=-1,
                verbosity=0)

        if use_gpu:
            self.params_bounds.update({'tree_method': 'gpu_hist', 'predictor': 'gpu_predictor'})

    def objective(self, **params):
        """重写目标函数"""

        # 纠正参数类型
        params = self._round_params(params)
        _params = {**self.params_bounds, **params}

        # 核心逻辑
        rst = xgb.cv(_params,
                     self.data,
                     num_boost_round=10000,
                     nfold=self.cv,
                     early_stopping_rounds=300,
                     verbose_eval=0,
                     show_stdv=False,
                     as_pandas=False,
                     stratified=True,
                     shuffle=True,
                     seed=self.cv_seed)

        scores = rst['test-auc-mean']
        self.best_iterations.append(len(scores))
        return scores[-1]

    @property
    def res(self):
        res = self.optimizer.res.copy()
        for res_, iteration in zip(res, self.best_iterations):
            _ = {**self.params_bounds, 'n_estimators': iteration}
            res_['params'] = self._round_params(_)
        return res

    @property
    def best_params(self):
        return self.res[self.optimizer._space.target.argmax()]

    def to_file(self, file='opt', mode='json'):
        """

        :param mode: {'json', 'yml'}
        :return:
        """
        with open(f"{file}.{mode}", 'w') as f:
            if mode == 'json':
                json.dump(self.res, f, indent=4)
            else:
                yaml.safe_dump(eval(str(self.res)), f)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : tql-Python.
# @File         : demo
# @Time         : 2019-09-16 15:23
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 
"""
https://www.cnblogs.com/zhaoyingjie/p/9664081.html
"""

from datetime import datetime
import os

from apscheduler.schedulers.blocking import BlockingScheduler

scheduler = BlockingScheduler()


@scheduler.scheduled_job('interval', seconds=3)
def tick():
    print('Tick! The time is: %s' % datetime.now())


scheduler.start()

# if __name__ == '__main__':
#
#     scheduler.add_job(tick, 'interval', seconds=3)
#     print('Press Ctrl+{0} to exit'.format('Break' if os.name == 'nt' else 'C'))
#
#     try:
#         scheduler.start()
#     except (KeyboardInterrupt, SystemExit):
#         pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : tql-Python.
# @File         : params
# @Time         : 2019-09-25 17:05
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 


params_lgb = {
    'n_estimators': 10000,
    'metric': 'auc',
    'boosting_type': 'gbdt',
    'objective': 'binary',
    'max_depth': -1,
    'num_leaves': 16,
    'learning_rate': 0.005,

    'subsample': 0.788,
    'subsample_freq': 8,
    'colsample_bytree': 0.617,

    'reg_alpha': 0.631,
    'reg_lambda': 0.81,
    'min_split_gain': 0.884,
    'min_child_weight': 0.01,
    'min_child_samples': 31,

    'n_jobs': -1,
    'verbosity': -1,
    'scale_pos_weight': 1,

    'random_state': 666,
    'bagging_seed': 0,
    'feature_fraction_seed': 0,
}

params_xgb = {
    'n_estimators': 10000,
    'eval_metric': 'auc',
    'booster': 'gbtree',
    'objective': 'binary:logistic',
    'max_depth': 7,
    'learning_rate': 0.02421306293966501,

    'subsample': 0.5795222896954795,
    'colsample_bytree': 0.19985830478870198,
    'colsample_bylevel': 1,

    'reg_alpha': 5.455077557037526,
    'reg_lambda': 5.8700819100907715,

    'gamma': 0.2484205604663352,
    'min_child_weight': 19.75111813614344,
    'max_delta_step': 1.4546423084703382,

    'n_jobs': -1,
    'verbosity': 0,
    'scale_pos_weight': 1,

    'tree_method': 'gpu_hist',
    'predictor': 'gpu_predictor',
    'random_state': 666
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : tql-Python.
# @File         : demo
# @Time         : 2019-09-16 14:17
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  :


from tql.algo_ml.opt_params import LGBMOptimizer, XGBOptimizer, CatBoostOptimizer, Optimizer

from sklearn.datasets import make_classification

X, y = make_classification(1000)

opt = Optimizer(X, y, cv=5, cv_seed=0)
print(opt.maximize(1))

opt = LGBMOptimizer(X, y, cv=5, cv_seed=0)
print(opt.maximize(1))

opt = XGBOptimizer(X, y, cv=5, cv_seed=0, use_gpu=0)
print(opt.maximize(1))

opt = CatBoostOptimizer(X, y, cv=5, cv_seed=0, use_gpu=0)
print(opt.maximize(1))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : tql-Python.
# @File         : keras
# @Time         : 2019-09-27 14:23
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 
"""https://stackoverflow.com/questions/43368241/student-teacher-model-in-keras/43700164"""

import keras
from keras.datasets import mnist
from keras.layers import Input, Embedding, LSTM, Dense, Lambda
from keras.models import Model
import numpy as np
from keras.utils import np_utils
from keras.layers.core import Dense, Dropout, Activation
from keras.models import Sequential
from keras.layers import Dense, Merge
from keras.optimizers import SGD, Adam, RMSprop
from keras.layers import *

nb_classes = 10

(X_train, y_train), (X_test, y_test) = mnist.load_data()

X_train = X_train.reshape(60000, 784)
X_test = X_test.reshape(10000, 784)
X_train = X_train.astype('float32')
X_test = X_test.astype('float32')
X_train /= 255
X_test /= 255
print(X_train.shape[0], 'train samples')
print(X_test.shape[0], 'test samples')

# convert class vectors to binary class matrices
Y_train = np_utils.to_categorical(y_train, nb_classes)
Y_test = np_utils.to_categorical(y_test, nb_classes)



batch_size = 128
nb_classes = 10
nb_epoch = 3

teacher = Sequential()
teacher.add(Dense(10, input_shape=(784,)))
teacher.add(Dense(10))
teacher.add(Activation('softmax'))

teacher.summary()
teacher.compile(loss='categorical_crossentropy',
                optimizer=RMSprop(),
                metrics=['accuracy'])

history = teacher.fit(X_train, Y_train,
                      batch_size=batch_size, nb_epoch=nb_epoch,
                      verbose=1, validation_data=(X_test, Y_test))
score = teacher.evaluate(X_test, Y_test, verbose=0)
print('Test score:', score[0])
print('Test accuracy:', score[1])

for i in range(len(teacher.layers)):
    setattr(teacher.layers[i], 'trainable', False)

Y_train = np.zeros((60000, 10))

student = Sequential()
student.add(Dense(10, input_dim=784))
student.add(Activation('softmax'))
student.compile(loss='mean_squared_error', optimizer='Adam', metrics=['accuracy'])



def negativeActivation(x):
    return -x


negativeRight = Activation(negativeActivation)(student.output)
diff = Add()([teacher.output, negativeRight])

model = Model(inputs=[teacher.input, student.input], outputs=[diff])
model.compile(loss='mean_squared_error', optimizer='Adam', metrics=['acc'])

model.summary(line_length=150)
model.fit([X_train, X_train], [Y_train], batch_size=128, nb_epoch=5)

print
student.evaluate(X_test, Y_test)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : tql-Python.
# @File         : BaseOOF
# @Time         : 2019-06-23 20:07
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  :

import time
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from sklearn.model_selection import StratifiedKFold, RepeatedStratifiedKFold, KFold, cross_val_predict, cross_validate
from sklearn.metrics import roc_auc_score

from xgboost import XGBClassifier


class XGBClassifierCV(object):
    """cross_val_predict"""

    def __init__(self, params=None, cv=5, random_state=None, n_repeats=None):
        self.clf = XGBClassifier()
        self.cv = cv
        if params:
            self.clf.set_params(**params)
        if n_repeats:
            self._kf = RepeatedStratifiedKFold(cv, True, random_state)
            self._num_preds = cv * n_repeats
        else:
            self._kf = StratifiedKFold(cv, True, random_state)
            self._num_preds = cv

    def fit(self, X, y, X_test=None, feval=roc_auc_score, sample_weight=None, eval_metric='auc',
            early_stopping_rounds=300, verbose=100, xgb_model=None, sample_weight_eval_set=None, callbacks=None):
        """输入数组"""
        self.best_info = {}
        if X_test is None:
            X_test = X[:1]

        self.oof_train = np.zeros(len(X))
        self.oof_test = np.zeros((len(X_test), self._num_preds))
        for n_fold, (train_index, valid_index) in enumerate(self._kf.split(X, y)):
            if verbose:
                print("\033[94mFold %s started at %s\033[0m" % (n_fold + 1, time.ctime()))
            X_train, y_train = X[train_index], y[train_index]
            X_valid, y_valid = X[valid_index], y[valid_index]
            eval_set = [(X_train, y_train), (X_valid, y_valid)]

            ########################################################################
            self.clf.fit(X_train, y_train, sample_weight, eval_set, eval_metric, early_stopping_rounds, verbose,
                         xgb_model, sample_weight_eval_set, callbacks)

            self.oof_train[valid_index] = self.clf.predict_proba(X_valid)[:, 1]
            self.oof_test[:, n_fold] = self.clf.predict_proba(X_test)[:, 1]

            # best info
            # print(self.clf.evals_result_)
            self.best_info.setdefault('best_iteration', []).append(self.clf.best_iteration)
            self.best_info.setdefault('best_score_valid', []).append(self.clf.best_score)
            ########################################################################

        # 输出 测试集 oof
        self.oof_test_rank = (pd.DataFrame(self.oof_test).rank().mean(1) / len(self.oof_test)).values
        self.oof_test = self.oof_test.mean(1)

        assert len(X) == len(self.oof_train)
        assert len(X_test) == len(self.oof_test)

        # 计算 训练集 oof 得分
        if feval:
            self.oof_score = feval(y, self.oof_train)
            print(f"\n\033[94mScore Info:\033[0m")
            print(f"\033[94m     {self.cv:>2} CV: {self.oof_score:.6f}\033[0m")

            _ = np.array(self.best_info['best_iteration'])
            print(f"\033[94m      Iter: {_.mean():.0f} +/- {_.std():.0f}\033[0m")

            _ = np.array(self.best_info['best_score_valid'])
            print(f"\033[94m     Valid: {_.mean():.6f} +/- {_.std():.6f} \033[0m\n")

        return self.oof_score

    def oof_submit(self, ids, pred_ranking=False, file=None, preds=None):
        """preds藏分用"""
        if file is None:
            file = f'submit_cv{self.cv}_{self.oof_score}.csv'
        print(f'Save {file} ...')

        if preds is None:
            preds = self.oof_test_rank if pred_ranking else self.oof_test

        if not isinstance(ids, pd.DataFrame):
            ids = pd.DataFrame(ids)
        ids.assign(preds=preds).to_csv(file, index=False, header=False)

    @property
    def oof_train_and_test(self):
        return np.r_[self.oof_train, self.oof_test]

    def oof_save(self, file='./oof_train_and_test.csv'):
        pd.DataFrame(self.oof_train_and_test, columns=['oof_train_and_test']).to_csv(file, index=False)

    def plot_feature_importances(self, feature_names=None, topk=20, figsize=(10, 6), pic_name=None):
        columns = ['Importances', 'Features']
        importances = self.clf.feature_importances_.tolist()
        if feature_names is None:
            feature_names = list(map(lambda x: f'F_{x}', range(len(importances))))
        _ = list(zip(importances, feature_names))
        df = pd.DataFrame(_, columns=columns).sort_values('Importances', 0, False)

        plt.figure(figsize=figsize)
        sns.barplot(*columns, data=df[:topk])
        plt.title('Features Importances\n')
        plt.tight_layout()
        if pic_name is None:
            plt.savefig(f'importances_{self.oof_score}.png')


if __name__ == "__main__":
    from sklearn.datasets import make_classification

    X, y = make_classification()
    X_test, _ = make_classification()

    clf = XGBClassifierCV({'n_estimators': 1000})
    clf.fit(X, y, X_test)
